//////////////////////////////////////////////////////////////////////
// app.cpp
//
// "Uninteresting" portion of the rt_FrameBuffer sample consisting of
// application boiler plate code and GUI handling.
//
// Rich Thomson
// Copyright (C) 2002, Pahvant Technologies, all rights reserved.
//
#include <sstream>

#include <d3dx8.h>

#include "D3DUtil.h"
#include "DXUtil.h"

#include "colorsel.h"
#include "resource.h"
#include "rt_FrameBuffer.h"

#include "rt/app.h"
#include "rt/hr.h"
#include "rt/mat.h"
#include "rt/mesh.h"
#include "rt/misc.h"
#include "rt/surface.h"
#include "rt/texture.h"
#include "rt/vertexbuf.h"

//////////////////////////////////////////////////////////////////////
// background tile and stipple size
//
const UINT TILE_SIZE = 64;
const UINT STIPPLE_SIZE = 64;

//////////////////////////////////////////////////////////////////////
// s_screen_vertex::FVF
//
// FVF code for screen-space vertices used to draw the stipple texture
//
const DWORD s_screen_vertex::FVF =
    D3DFVF_XYZRHW | D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE2(0);

//////////////////////////////////////////////////////////////////////
// g_jitterN, g_jitter
//
// jitter coefficients used for depth of field multisampling
//
const float g_jitter1[2] =
{
    0, 0
};
const float g_jitter2[4] =
{
    0.25f,          0.75f,
    0.75f,          0.25f
};
const float g_jitter3[6] =
{
    0.5033922635f,  0.8317967229f,
    0.7806016275f,  0.2504380877f,
    0.2261828938f,  0.4131553612f
};
const float g_jitter4[8] =
{
    0.375f,         0.25f,
    0.125f,         0.75f,
    0.875f,         0.25f,
    0.625f,         0.75f
};
const float g_jitter5[10] =
{
    0.5f,           0.5f,
    0.3f,           0.1f,
    0.7f,           0.9f,
    0.9f,           0.3f,
    0.1f,           0.7f
};
const float g_jitter6[12] =
{
    0.4646464646f,  0.4646464646f,
    0.1313131313f,  0.7979797979f,
    0.5353535353f,  0.8686868686f,
    0.8686868686f,  0.5353535353f,
    0.7979797979f,  0.1313131313f, 
    0.2020202020f,  0.2020202020f
};
const float g_jitter8[16] =
{
    0.5625f,        0.4375f,
    0.0625f,        0.9375f,
    0.3125f,        0.6875f,
    0.6875f,        0.8125f,
    0.8125f,        0.1875f,
    0.9375f,        0.5625f,
    0.4375f,        0.0625f,
    0.1875f,        0.3125f
};
const float g_jitter9[18] =
{
    0.5f,           0.5f,
    0.1666666666f,  0.944444444f,
    0.5f,           0.1666666666f,
    0.5f,           0.8333333333f,
    0.1666666666f,  0.2777777777f,
    0.8333333333f,  0.3888888888f,
    0.1666666666f,  0.6111111111f,
    0.8333333333f,  0.7222222222f,
    0.8333333333f,  0.0555555555f,
};
const float g_jitter12[24] =
{
    0.4166666666f,  0.625f,
    0.9166666666f,  0.875f,
    0.25f,          0.375f,
    0.4166666666f,  0.125f,
    0.75f,          0.125f,
    0.0833333333f,  0.125f,
    0.75f,          0.625f,
    0.25f,          0.875f,
    0.5833333333f,  0.375f,
    0.9166666666f,  0.375f,
    0.0833333333f,  0.625f,
    0.5833333333f,  0.875f
};
const float g_jitter16[32] =
{
    0.375f,         0.4375f,
    0.625f,         0.0625f,
    0.875f,         0.1875f,
    0.125f,         0.0625f,
    0.375f,         0.6875f,
    0.875f,         0.4375f,
    0.625f,         0.5625f,
    0.375f,         0.9375f,
    0.625f,         0.3125f,
    0.125f,         0.5625f,
    0.125f,         0.8125f,
    0.375f,         0.1875f,
    0.875f,         0.9375f,
    0.875f,         0.6875f,
    0.125f,         0.3125f,
    0.625f,         0.8125f
};
const float *g_jitter[16] =
{
    g_jitter1,
    g_jitter2,
    g_jitter3,
    g_jitter4,
    g_jitter5,
    g_jitter6,
    g_jitter8,
    g_jitter8,
    g_jitter9,
    g_jitter12,
    g_jitter12,
    g_jitter12,
    g_jitter16,
    g_jitter16,
    g_jitter16,
    g_jitter16
};

//-----------------------------------------------------------------------------
// Global access to the app (needed for the global WndProc())
//-----------------------------------------------------------------------------
CMyD3DApplication* g_pApp  = NULL;
HINSTANCE          g_hInst = NULL;

//////////////////////////////////////////////////////////////////////
// parse_command_line
//
// Parse the command line, looking for the switches:
//
//  -nodest     Don't use destination alpha
//  -nostencil  Don't use stencil planes
//
// Throw up a message box if we encounter something we don't recognize.
//
bool
parse_command_line(LPCTSTR command_line,
                   BOOL &destination_alpha, DWORD &stencil_bits)
{
    rt::tistringstream args(command_line);
    rt::tstring arg;
    while (!args.eof())
    {
        args >> arg;
        if (_T("-nodest") == arg)
        {
            destination_alpha = FALSE;
        }
        else if (_T("-nostencil") == arg)
        {
            stencil_bits = 0;
        }
        else if (arg.length())
        {
            rt::tostringstream barf;
            barf << _T("I did not recognize the switch '") << arg
                << _T("' in the command line\n\"") << command_line
                << _T("\".\n\nValid switches are:\n")
                << _T("-nodest\t\tDon't use destination alpha.\n")
                << _T("-nostencil\t\tDon't use stencil planes.\n");
            ::MessageBox(NULL, barf.str().c_str(),
                _T("rt_FrameBuffer: Invalid Command Line Argument"), MB_OK);

            return false;
        }
    }

    return true;
}

//-----------------------------------------------------------------------------
// Name: _tWinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI
_tWinMain(HINSTANCE instance, HINSTANCE, LPTSTR command_line, int)
{
    BOOL destination_alpha = TRUE;
    DWORD stencil_bits = 2;
    if (!parse_command_line(command_line, destination_alpha, stencil_bits))
    {
        return 0;
    }

    CMyD3DApplication d3dApp(destination_alpha, stencil_bits);

    g_pApp  = &d3dApp;
    g_hInst = instance;

    if (FAILED(d3dApp.Create(instance)))
        return 0;

    return d3dApp.Run();
}




//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication(BOOL destination_alpha,
                                     DWORD stencil_bits)
    : m_bLoadingApp(TRUE),
    m_pFont(new CD3DFont(_T("Arial"), 10, D3DFONT_BOLD)),
    m_UserInput(),
    m_fWorldRotX(0.0f),
    m_fWorldRotY(0.0f),
    m_last_rot_x(0.0f),
    m_last_rot_y(0.0f),
    m_teapot(),
    m_texture(),
    m_tile(),
    m_device_tile(),
    m_torus(),
    m_stipple_verts(),
    m_stipple(),
    m_tile_rects(),
    m_tile_offsets(),
    m_texture_file(_T("flag.bmp")),
    m_num_stipple_quads(0),
    m_text_fg(D3DCOLOR_XRGB(255, 255, 0)),
    m_specular(D3DCOLOR_XRGB(255, 255, 255)),
    m_bg(D3DCOLOR_XRGB(0, 0, 0)),
    m_fg(D3DCOLOR_XRGB(255, 255, 255)),
    m_light_color(D3DCOLOR_XRGB(255, 255, 255)),
    m_blending(E_BLENDING_NONE),
    m_multisample_motion_blur(false),
    m_multisample_depth_of_field(false),
    m_multisample_antialias(false),
    m_color_write_enable_red(true),
    m_color_write_enable_green(true),
    m_color_write_enable_blue(true),
    m_color_write_enable_alpha(true),
    m_z_write_enable(true),
    m_textured(false),
    m_stencil_stippling(false),
    m_stencil_mask(false),
    m_show_stats(true),
    m_animate_view(false),
    m_dithered(false),
    m_can_texture(false),
    m_destination_alpha(false),
    m_tile_background(false),
    m_build_stencil(false)
{
    m_dwCreationWidth = 500;
    m_dwCreationHeight = 375;
    m_strWindowTitle = TEXT("rt_FrameBuffer");

    // we need a depth buffer
    m_bUseDepthBuffer = TRUE;

    // we need destination alpha for blending
    m_bDestinationAlpha = destination_alpha;

    // we need two stencil planes: 1 for masking, 1 for stippling
    m_dwMinStencilBits = stencil_bits;

    // Read settings from registry
    ReadSettings();
}




//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
    // TODO: perform one time initialization

    // Drawing loading status message until app finishes loading
    SendMessage( m_hWnd, WM_PAINT, 0, 0 );

    m_bLoadingApp = FALSE;

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: ReadSettings()
// Desc: Read the app settings from the registry
//-----------------------------------------------------------------------------
VOID CMyD3DApplication::ReadSettings()
{
    HKEY hkey;
    if( ERROR_SUCCESS == RegCreateKeyEx( HKEY_CURRENT_USER, DXAPP_KEY,
        0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hkey, NULL ) )
    {
        // TODO: change as needed

        // Read the stored window width/height.  This is just an example,
        // of how to use DXUtil_Read*() functions.
        DXUtil_ReadIntRegKey( hkey, TEXT("Width"), &m_dwCreationWidth, m_dwCreationWidth );
        DXUtil_ReadIntRegKey( hkey, TEXT("Height"), &m_dwCreationHeight, m_dwCreationHeight );

        RegCloseKey( hkey );
    }
}




//-----------------------------------------------------------------------------
// Name: WriteSettings()
// Desc: Write the app settings to the registry
//-----------------------------------------------------------------------------
VOID CMyD3DApplication::WriteSettings()
{
    HKEY hkey;
    DWORD dwType = REG_DWORD;
    DWORD dwLength = sizeof(DWORD);

    if( ERROR_SUCCESS == RegCreateKeyEx( HKEY_CURRENT_USER, DXAPP_KEY,
        0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hkey, NULL ) )
    {
        // TODO: change as needed

        // Write the window width/height.  This is just an example,
        // of how to use DXUtil_Write*() functions.
        DXUtil_WriteIntRegKey( hkey, TEXT("Width"), m_rcWindowClient.right );
        DXUtil_WriteIntRegKey( hkey, TEXT("Height"), m_rcWindowClient.bottom );

        RegCloseKey( hkey );
    }
}





//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the display device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{
    bool use_ref = false;

    if (use_ref)
    {
        return (pCaps->DeviceType == D3DDEVTYPE_REF) ? S_OK : E_FAIL;
    }

    BOOL bCapsAcceptable;

    // TODO: Perform checks to see if these display caps are acceptable.
    bCapsAcceptable = TRUE;

    if( bCapsAcceptable )
        return S_OK;
    else
        return E_FAIL;
}

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
    // rather than reject devices without this cap, we simply disable
    // texturing and allow rendering with the diffuse color on this device.
    m_can_texture =
             (m_d3dCaps.TextureFilterCaps & D3DPTFILTERCAPS_MINFLINEAR) &&
             (m_d3dCaps.TextureFilterCaps & D3DPTFILTERCAPS_MAGFLINEAR) &&
             (m_d3dCaps.TextureFilterCaps & D3DPTFILTERCAPS_MIPFLINEAR) &&
             (m_d3dCaps.TextureOpCaps & D3DTEXOPCAPS_SELECTARG2) &&
             (m_d3dCaps.TextureOpCaps & D3DTEXOPCAPS_MODULATE) &&
             (m_d3dCaps.TextureAddressCaps & D3DPTADDRESSCAPS_WRAP) &&
             (m_d3dCaps.VertexProcessingCaps & D3DVTXPCAPS_TEXGEN);
    if (m_can_texture)
    {
        create_texture();
        {
            DWORD passes = 0;
            const HRESULT hr = m_pd3dDevice->ValidateDevice(&passes);
            if (FAILED(hr))
            {
                m_can_texture = false;
            }
        }
    }
    if (!m_can_texture)
    {
        m_textured = false;
        m_texture = 0;
    }

    // Init the font
    THR(m_pFont->InitDeviceObjects(m_pd3dDevice));

    // Create a teapot mesh using D3DX and optimize it
    {
        rt::dx_buffer<DWORD> adj;
        THR(::D3DXCreateTeapot(m_pd3dDevice, &m_teapot, &adj));
        std::vector<DWORD> adj_out(m_teapot->GetNumFaces()*3);
        std::vector<DWORD> face_remap(m_teapot->GetNumFaces());
        CComPtr<ID3DXBuffer> vertex_remap;
        THR(m_teapot->OptimizeInplace(D3DXMESHOPT_ATTRSORT, adj,
            &adj_out[0], &face_remap[0], &vertex_remap));
    }

    // create a background ARGB tile with three steps in alpha in Y
    // and a gradient in rgb in X
    THR(m_pd3dDevice->CreateImageSurface(TILE_SIZE, TILE_SIZE,
        D3DFMT_A8R8G8B8, &m_tile));
    {
        rt::surface_lock lock(m_tile);
        for (UINT y = 0; y < TILE_SIZE; y++)
        {
            DWORD *row = lock.scanline32(y);
            BYTE alpha = 0;
            if (y < TILE_SIZE/3)
            {
                alpha = 255/3;
            }
            else if (y < 2*TILE_SIZE/3)
            {
                alpha = 255*2/3;
            }
            else
            {
                alpha = 255;
            }
            for (UINT x = 0; x < TILE_SIZE; x++)
            {
                const BYTE rgb = alpha*x/(TILE_SIZE-1);
                row[x] = D3DCOLOR_ARGB(alpha, rgb, rgb, rgb);
            }
        }
    }

    // create a torus used to create the stencil mask
    THR(::D3DXCreateTorus(m_pd3dDevice, 0.325f, 0.65f, 10, 50, &m_torus, NULL));

    // create a stipple texture pattern
    THR(::D3DXCreateTexture(m_pd3dDevice, STIPPLE_SIZE, STIPPLE_SIZE, 1, 0,
        D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &m_stipple));
    {
        rt::texture_lock lock(m_stipple);
        for (UINT y = 0; y < STIPPLE_SIZE; y++)
        {
            DWORD *texels = lock.scanline32(y);
            for (UINT x = 0; x < STIPPLE_SIZE; x++)
            {
                if ((y/4) & 1)
                {
                    texels[x] = ((x/4) & 1) ? ~0 : 0;
                }
                else
                {
                    texels[x] = ((x/4) & 1) ? 0 : ~0;
                }
            }
        }
    }

    return S_OK;
}


//////////////////////////////////////////////////////////////////////
// enable_menu
//
// Enable or disable a menu item based on a boolean.
//
inline void
enable_menu(HMENU menu, UINT item, bool enabled)
{
    ::EnableMenuItem(menu, item, enabled ? MF_ENABLED : MF_GRAYED);
}

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restores scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
    HMENU menu = ::GetMenu(m_hWnd);
    // if D3DRS_COLORWRITEENABLE is not supported, disable menu items
    const bool mask_colors =
        (m_d3dCaps.PrimitiveMiscCaps & D3DPMISCCAPS_COLORWRITEENABLE) != 0;
    ::enable_menu(menu, IDM_COLOR_WRITE_ENABLE_RED, mask_colors);
    ::enable_menu(menu, IDM_COLOR_WRITE_ENABLE_GREEN, mask_colors);
    ::enable_menu(menu, IDM_COLOR_WRITE_ENABLE_BLUE, mask_colors);
    ::enable_menu(menu, IDM_COLOR_WRITE_ENABLE_ALPHA, mask_colors);
    m_destination_alpha = (D3DFMT_A8R8G8B8 == m_d3dsdBackBuffer.Format) ||
        (D3DFMT_A1R5G5B5 == m_d3dsdBackBuffer.Format);
    // toggle menu items based on destination alpha supported
    ::enable_menu(menu, IDM_BLEND_DEST, m_destination_alpha);
    ::enable_menu(menu, IDM_BLEND_DEST_OVER_SRC, m_destination_alpha);
    ::enable_menu(menu, IDM_BLEND_SRC_IN_DEST, m_destination_alpha);
    ::enable_menu(menu, IDM_BLEND_SRC_ATOP_DEST, m_destination_alpha);
    ::enable_menu(menu, IDM_BLEND_DEST_ATOP_SRC, m_destination_alpha);
    ::enable_menu(menu, IDM_BLEND_SRC_OUT_DEST, m_destination_alpha);
    ::enable_menu(menu, IDM_BLEND_SRC_XOR_DEST, m_destination_alpha);

    // if we don't have destination alpha, we might have to change the
    // blending selection
    if (!m_destination_alpha)
    {
        switch (m_blending)
        {
        case E_BLENDING_DEST:
        case E_BLENDING_DEST_OVER_SRC:
        case E_BLENDING_SRC_IN_DEST:
        case E_BLENDING_SRC_ATOP_DEST:
        case E_BLENDING_DEST_ATOP_SRC:
        case E_BLENDING_SRC_OUT_DEST:
        case E_BLENDING_SRC_XOR_DEST:
            rt::check_menu(menu, IDM_BLEND_CLEAR + m_blending, false);
            m_blending = E_BLENDING_NONE;
            rt::check_menu(menu, IDM_BLEND_CLEAR + m_blending, true);
            break;
        }
    }
    
    const DWORD blend_caps[] =
    {
        0,
        D3DPBLENDCAPS_ZERO,
        D3DPBLENDCAPS_ONE,
        D3DPBLENDCAPS_SRCCOLOR,
        D3DPBLENDCAPS_INVSRCCOLOR,
        D3DPBLENDCAPS_SRCALPHA,
        D3DPBLENDCAPS_INVSRCALPHA,
        D3DPBLENDCAPS_DESTALPHA,
        D3DPBLENDCAPS_INVDESTALPHA,
        D3DPBLENDCAPS_DESTCOLOR,
        D3DPBLENDCAPS_INVDESTCOLOR
    };
    for (UINT i = 0; i < NUM_OF(g_blend_factors); i++)
    {
        // blend factors used in this blending operator are not supported
        if (!(m_d3dCaps.SrcBlendCaps & blend_caps[g_blend_factors[i].m_src])
            || !(m_d3dCaps.DestBlendCaps & blend_caps[g_blend_factors[i].m_dest]))
        {
            // unsupported blending operator selected?
            if (m_blending == i)
            {
                // change the blending to none
                rt::check_menu(menu, IDM_BLEND_CLEAR + m_blending, false);
                m_blending = E_BLENDING_NONE;
                rt::check_menu(menu, IDM_BLEND_CLEAR + m_blending, true);
            }

            // disable menu item for blending operator i
            ::enable_menu(menu, IDM_BLEND_CLEAR + i, false);
        }
        else
        {
            ::enable_menu(menu, IDM_BLEND_CLEAR + i, true);
        }
    }

    // can't do any stencil effects without stencil bits
    if (m_dwMinStencilBits < 2)
    {
        m_stencil_mask = false;
        m_stencil_stippling = false;
        rt::check_menu(menu, IDM_STENCIL_IRREGULAR_MASK, false);
        rt::check_menu(menu, IDM_STENCIL_STIPPLING, false);
        ::enable_menu(menu, IDM_STENCIL_IRREGULAR_MASK, false);
        ::enable_menu(menu, IDM_STENCIL_STIPPLING, false);
    }

    // can't do any of these without multisampling
    if (D3DMULTISAMPLE_NONE == m_d3dpp.MultiSampleType)
    {
        m_multisample_antialias = false;
        m_multisample_depth_of_field = false;
        m_multisample_motion_blur = false;

        rt::check_menu(menu, IDM_MULTISAMPLE_ANTIALIAS, false);
        rt::check_menu(menu, IDM_MULTISAMPLE_DEPTH_OF_FIELD, false);
        rt::check_menu(menu, IDM_MULTISAMPLE_MOTION_BLUR, false);

        ::enable_menu(menu, IDM_MULTISAMPLE_ANTIALIAS, false);
        ::enable_menu(menu, IDM_MULTISAMPLE_DEPTH_OF_FIELD, false);
        ::enable_menu(menu, IDM_MULTISAMPLE_MOTION_BLUR, false);
    }
    // if this cap is set, can't do motion blur or depth of field
    else if (m_d3dCaps.RasterCaps & D3DPRASTERCAPS_STRETCHBLTMULTISAMPLE)
    {
        m_multisample_depth_of_field = false;
        m_multisample_motion_blur = false;

        rt::check_menu(menu, IDM_MULTISAMPLE_DEPTH_OF_FIELD, false);
        rt::check_menu(menu, IDM_MULTISAMPLE_MOTION_BLUR, false);

        ::enable_menu(menu, IDM_MULTISAMPLE_ANTIALIAS, true);
        ::enable_menu(menu, IDM_MULTISAMPLE_DEPTH_OF_FIELD, false);
        ::enable_menu(menu, IDM_MULTISAMPLE_MOTION_BLUR, false);
    }
    else
    {
        ::enable_menu(menu, IDM_MULTISAMPLE_ANTIALIAS, true);
        ::enable_menu(menu, IDM_MULTISAMPLE_DEPTH_OF_FIELD, true);
        ::enable_menu(menu, IDM_MULTISAMPLE_MOTION_BLUR, true);
    }

    // Set up our view matrix. A view matrix can be defined given an eye point,
    // a point to lookat, and a direction for which way is up. Here, we set the
    // eye five units back along the z-axis and up three units, look at the
    // origin, and define "up" to be in the y-direction.
    rt::mat_look_at view(D3DXVECTOR3(0, 0, -5));
    THR(m_pd3dDevice->SetTransform(D3DTS_VIEW, &view));

    // Set the projection matrix
    D3DXMATRIX matProj;
    FLOAT fAspect = ((FLOAT)m_d3dsdBackBuffer.Width) / m_d3dsdBackBuffer.Height;
    D3DXMatrixPerspectiveFovLH( &matProj, D3DX_PI/4, fAspect, 1.0f, 7.0f );
    m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );

    // Set up lighting states
    D3DLIGHT8 light;
    ::D3DUtil_InitLight(light, D3DLIGHT_DIRECTIONAL, -1.0f, -1.0f, 2.0f);
    m_pd3dDevice->SetLight(0, &light);
    m_pd3dDevice->LightEnable(0, TRUE);
    m_pd3dDevice->SetRenderState(D3DRS_LIGHTING, TRUE);

    // Restore the font
    m_pFont->RestoreDeviceObjects();

    // create a tile in the same format as the back buffer and load it
    THR(m_pd3dDevice->CreateImageSurface(TILE_SIZE, TILE_SIZE,
        m_d3dsdBackBuffer.Format, &m_device_tile));
    THR(::D3DXLoadSurfaceFromSurface(m_device_tile, NULL, NULL,
        m_tile, NULL, NULL, D3DX_FILTER_NONE, 0));

    // rebuild stencil planes if needed
    m_build_stencil = m_stencil_mask || m_stencil_stippling;

    // build RECTs and POINTs for tiling the background
    m_tile_rects.clear();
    m_tile_offsets.clear();
    const UINT width = m_d3dsdBackBuffer.Width;
    const UINT height = m_d3dsdBackBuffer.Height;
    UINT y;
    for (y = 0; y < height; y += 2*TILE_SIZE)
    {
        for (UINT x = 0; x < width; x += 2*TILE_SIZE)
        {
            const POINT offset = { x, y };
            const RECT rect =
            {
                0, 0,
                (x + TILE_SIZE) <= width ? TILE_SIZE : width-x,
                (y + TILE_SIZE) <= height ? TILE_SIZE : height-y
            };
            m_tile_rects.push_back(rect);
            m_tile_offsets.push_back(offset);
        }
    }

    // build s_screen_vertex structures for stippling the window
    m_stipple_verts = 0;
    m_num_stipple_quads = (width + STIPPLE_SIZE - 1)/STIPPLE_SIZE*
        (height + STIPPLE_SIZE - 1)/STIPPLE_SIZE;
    THR(m_pd3dDevice->CreateVertexBuffer(m_num_stipple_quads*6*sizeof(s_screen_vertex),
        D3DUSAGE_WRITEONLY, s_screen_vertex::FVF, D3DPOOL_MANAGED,
        &m_stipple_verts));
    rt::vertex_lock<s_screen_vertex> verts(m_stipple_verts);
    s_screen_vertex *quad = verts.data();
    for (y = 0; y < height; y += STIPPLE_SIZE)
    {
        for (UINT x = 0; x < width; x += STIPPLE_SIZE)
        {
            const float right = min(x + STIPPLE_SIZE, width);
            const float bottom = min(y + STIPPLE_SIZE, height);
            quad[0].x = x + 0.5f;
            quad[0].y = y + 0.5f;
            quad[0].z = 0.99f;
            quad[0].rhw = 0.99f;
            quad[0].u = 0;
            quad[0].v = 1;

            quad[1].x = right + 0.5f;
            quad[1].y = y + 0.5f;
            quad[1].z = 0.99f;
            quad[1].rhw = 0.99f;
            quad[1].u = 1;
            quad[1].v = 1;

            quad[2].x = x + 0.5f;
            quad[2].y = bottom + 0.5f;
            quad[2].z = 0.99f;
            quad[2].rhw = 0.99f;
            quad[2].u = 0;
            quad[2].v = 0;

            quad[3].x = x + 0.5f;
            quad[3].y = bottom + 0.5f;
            quad[3].z = 0.99f;
            quad[3].rhw = 0.99f;
            quad[3].u = 0;
            quad[3].v = 0;

            quad[4].x = right + 0.5f;
            quad[4].y = y + 0.5f;
            quad[4].z = 0.99f;
            quad[4].rhw = 0.99f;
            quad[4].u = 1;
            quad[4].v = 1;

            quad[5].x = right + 0.5f;
            quad[5].y = bottom + 0.5f;
            quad[5].z = 0.99f;
            quad[5].rhw = 0.99f;
            quad[5].u = 1;
            quad[5].v = 0;

            quad += 6;
        }
    }

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
    // Update user input state
    UpdateInput(&m_UserInput);

    m_last_rot_x = m_fWorldRotX;
    m_last_rot_y = m_fWorldRotY;

    // Update the world state according to user input
    if (m_animate_view)
    {
        m_fWorldRotY += m_fElapsedTime;
        if (m_fWorldRotY > 2*D3DX_PI)
        {
            m_fWorldRotY = fmodf(m_fWorldRotY, 2*D3DX_PI);
        }
    }
    else
    {
        if (m_UserInput.bRotateLeft && !m_UserInput.bRotateRight)
        {
            m_fWorldRotY += m_fElapsedTime;
        }
        else if (m_UserInput.bRotateRight && !m_UserInput.bRotateLeft)
        {
            m_fWorldRotY -= m_fElapsedTime;
        }
    }
    if (m_UserInput.bRotateUp && !m_UserInput.bRotateDown)
    {
        m_fWorldRotX += m_fElapsedTime;
    }
    else if (m_UserInput.bRotateDown && !m_UserInput.bRotateUp)
    {
        m_fWorldRotX -= m_fElapsedTime;
    }

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: UpdateInput()
// Desc: Update the user input.  Called once per frame
//-----------------------------------------------------------------------------
void CMyD3DApplication::UpdateInput( UserInput* pUserInput )
{
    pUserInput->bRotateUp    = ( m_bHasFocus && (GetAsyncKeyState( VK_UP )    & 0x8000) == 0x8000 );
    pUserInput->bRotateDown  = ( m_bHasFocus && (GetAsyncKeyState( VK_DOWN )  & 0x8000) == 0x8000 );
    pUserInput->bRotateLeft  = ( m_bHasFocus && (GetAsyncKeyState( VK_LEFT )  & 0x8000) == 0x8000 );
    pUserInput->bRotateRight = ( m_bHasFocus && (GetAsyncKeyState( VK_RIGHT ) & 0x8000) == 0x8000 );
}

//////////////////////////////////////////////////////////////////////
// operator<<(e_blending_operator)
//
// Formats an enumerant of e_blending_operator as a string.
//
rt::tostream &
operator<<(rt::tostream &stream, e_blending_operator blending)
{
    switch (blending)
    {
#define BLEND_OUT(enumerant_, str_) \
    case enumerant_: stream << _T(str_); break
        BLEND_OUT(E_BLENDING_CLEAR, "clear");
        BLEND_OUT(E_BLENDING_SRC, "src");
        BLEND_OUT(E_BLENDING_DEST, "dest");
        BLEND_OUT(E_BLENDING_SRC_OVER_DEST, "src over dest");
        BLEND_OUT(E_BLENDING_DEST_OVER_SRC, "dest over src");
        BLEND_OUT(E_BLENDING_SRC_IN_DEST, "src in dest");
        BLEND_OUT(E_BLENDING_DEST_IN_SRC, "dest in src");
        BLEND_OUT(E_BLENDING_SRC_ATOP_DEST, "src atop dest");
        BLEND_OUT(E_BLENDING_DEST_ATOP_SRC, "dest atop src");
        BLEND_OUT(E_BLENDING_SRC_OUT_DEST, "src out dest");
        BLEND_OUT(E_BLENDING_DEST_OUT_SRC, "dest out src");
        BLEND_OUT(E_BLENDING_SRC_XOR_DEST, "src xor dest");
        BLEND_OUT(E_BLENDING_NONE, "none");
#undef BLEND_OUT
    default:
        ATLASSERT(false);
    }
    return stream;
}


//-----------------------------------------------------------------------------
// Name: RenderText()
// Desc: Renders stats and help text to the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RenderText()
{
    // render text with a drop shadow
    render_text_aux(1, 1, D3DCOLOR_XRGB(0, 0, 0));
    render_text_aux(0, 0, m_text_fg);

    return S_OK;
}

//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::render_text_aux
//
// Render lines of text at a specific offset and in a specific color.
// Some program options are displayed in addition to the standard
// AppWizard sample text.
//
void
CMyD3DApplication::render_text_aux(float offset_x, float offset_y, D3DCOLOR fg)
{
    TCHAR szMsg[MAX_PATH] = TEXT("");

    // Output display stats
    FLOAT fNextLine = 60.0f + offset_y;
    float x = 2.0f + offset_x;

    {
        rt::tostringstream msg;
        msg << (m_destination_alpha ?
            _T("Destination alpha") : _T("No destination alpha"));

        if (m_dithered)
        {
            msg << _T(", dither");
        }

        if (E_BLENDING_NONE != m_blending)
        {
            msg << _T(", ") << m_blending;
        }

        if (m_multisample_antialias)
        {
            msg << _T(", multisample antialias");
        }
        else if (m_multisample_depth_of_field)
        {
            msg << _T(", multisample depth of field");
        }
        else if (m_multisample_motion_blur)
        {
            msg << _T(", multisample motion blur");
        }

        if (m_stencil_mask)
        {
            msg << _T(", stencil mask");
        }

        if (m_stencil_stippling)
        {
            msg << _T(", stencil stipple");
        }

        fNextLine -= 20.0f;
        m_pFont->DrawText(x, fNextLine, fg, msg.str().c_str());
    }

    lstrcpy( szMsg, m_strDeviceStats );
    fNextLine -= 20.0f;
    m_pFont->DrawText( x, fNextLine, fg, szMsg );

    lstrcpy( szMsg, m_strFrameStats );
    fNextLine -= 20.0f;
    m_pFont->DrawText( x, fNextLine, fg, szMsg );

    // Output statistics & help
    fNextLine = (FLOAT) m_d3dsdBackBuffer.Height + offset_y;

    wsprintf( szMsg, TEXT("Arrow keys: Up=%d Down=%d Left=%d Right=%d"),
              m_UserInput.bRotateUp, m_UserInput.bRotateDown, m_UserInput.bRotateLeft, m_UserInput.bRotateRight );
    fNextLine -= 20.0f; m_pFont->DrawText( x, fNextLine, fg, szMsg );

    lstrcpy( szMsg, TEXT("Use arrow keys to rotate object") );
    fNextLine -= 20.0f; m_pFont->DrawText( x, fNextLine, fg, szMsg );

    lstrcpy( szMsg, TEXT("Press 'F2' to configure display") );
    fNextLine -= 20.0f; m_pFont->DrawText( x, fNextLine, fg, szMsg );
}




//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Overrrides the main WndProc, so the sample can do custom message
//       handling (e.g. processing mouse, keyboard, or menu commands).
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc( HWND hWnd, UINT msg, WPARAM wParam,
                                    LPARAM lParam )
{
    switch( msg )
    {
    case WM_PAINT:
        if( m_bLoadingApp )
        {
            // Draw on the window tell the user that the app is loading
            // TODO: change as needed
            HDC hDC = GetDC( hWnd );
            TCHAR strMsg[MAX_PATH];
            wsprintf( strMsg, TEXT("Loading... Please wait") );
            RECT rct;
            GetClientRect( hWnd, &rct );
            DrawText( hDC, strMsg, -1, &rct, DT_CENTER|DT_VCENTER|DT_SINGLELINE );
            ReleaseDC( hWnd, hDC );
        }
        break;

    case WM_COMMAND:
        on_command(hWnd, wParam, lParam);
        break;
    }

    return CD3DApplication::MsgProc( hWnd, msg, wParam, lParam );
}




//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Invalidates device objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
    m_pFont->InvalidateDeviceObjects();
    m_device_tile = 0;
    m_stipple_verts = 0;
    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
    m_teapot = 0;
    m_torus = 0;
    m_texture = 0;
    m_tile = 0;
    m_device_tile = 0;
    m_stipple = 0;
    m_pFont->DeleteDeviceObjects();

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
    // TODO: Perform any final cleanup needed
    // Cleanup D3D font
    SAFE_DELETE( m_pFont );

    // Write the settings to the registry
    WriteSettings();

    return S_OK;
}

//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::on_command
//
// WM_COMMAND handler for this sample.  This mostly amounts to updating
// internal variables that reflect the current state of the UI, although
// some D3D state is updated directly from here.
//
void
CMyD3DApplication::on_command(HWND window, WPARAM wp, LPARAM lp)
{
    const HMENU menu = ::GetMenu(window);
    const UINT control = LOWORD(wp);

    switch (control)
    {
#define TOGGLE_COLOR_WRITE_ENABLE(id_, state_)                      \
    case IDM_COLOR_WRITE_ENABLE_##id_:                              \
        state_ = !state_;                                           \
        rt::check_menu(menu, IDM_COLOR_WRITE_ENABLE_##id_, state_); \
        set_color_write_enable();                                   \
        break
    TOGGLE_COLOR_WRITE_ENABLE(RED, m_color_write_enable_red);
    TOGGLE_COLOR_WRITE_ENABLE(GREEN, m_color_write_enable_green);
    TOGGLE_COLOR_WRITE_ENABLE(BLUE, m_color_write_enable_blue);
    TOGGLE_COLOR_WRITE_ENABLE(ALPHA, m_color_write_enable_alpha);
#undef TOGGLE_COLOR_WRITE_ENABLE
#define TOGGLE(id_, state_)                 \
    case id_:                               \
        state_ = !state_;                   \
        rt::check_menu(menu, id_, state_);  \
        break
    TOGGLE(IDM_Z_WRITE_ENABLE, m_z_write_enable);
    TOGGLE(IDM_ANIMATE_VIEW, m_animate_view);
    TOGGLE(IDM_SHOW_STATS, m_show_stats);
    TOGGLE(IDM_OPTION_DITHER, m_dithered);
    TOGGLE(IDM_OPTION_BACKGROUND_TILE, m_tile_background);
#undef TOGGLE

#define MODIFY_COLOR(id_, state_, transparent_) \
    case id_: \
        { \
            rt::pauser block(*this); \
            state_ = choose_color_transparent(window, state_, transparent_); \
        } \
        break
//    MODIFY_COLOR(IDM_EDIT_DIFFUSE_COLOR, m_fg, true);
    MODIFY_COLOR(IDM_EDIT_TEXT_COLOR, m_text_fg, true);
    MODIFY_COLOR(IDM_EDIT_SPECULAR_COLOR, m_specular, false);
#undef MODIFY_COLOR

    case IDM_EDIT_BG_COLOR:
        {
            rt::pauser block(*this);
            m_bg = choose_color(window, m_bg);
        }
        break;

    case IDM_EDIT_DIFFUSE_COLOR:
        {
            rt::pauser block(*this);
            m_fg = choose_color(window, m_fg);
        }
        break;

    case IDM_TEXTURED:
        m_textured = m_can_texture && !m_textured;
        rt::check_menu(menu, IDM_TEXTURED, m_textured);
        break;

    case IDM_FILE_OPEN_TEXTURE:
        {
            rt::pauser block(*this);
            on_texture_file(window, wp, lp);
        }
        break;

    case IDM_STENCIL_IRREGULAR_MASK:
        m_stencil_mask = !m_stencil_mask;
        rt::check_menu(menu, IDM_STENCIL_IRREGULAR_MASK, m_stencil_mask);
        m_build_stencil = true;
        break;

    case IDM_STENCIL_STIPPLING:
        m_stencil_stippling = !m_stencil_stippling;
        rt::check_menu(menu, IDM_STENCIL_STIPPLING, m_stencil_stippling);
        m_build_stencil = true;
        break;

    case IDM_MULTISAMPLE_ANTIALIAS:
        m_multisample_antialias = !m_multisample_antialias;
        m_multisample_depth_of_field = false;
        m_multisample_motion_blur = false;
        rt::check_menu(menu, IDM_MULTISAMPLE_ANTIALIAS, m_multisample_antialias);
        rt::check_menu(menu, IDM_MULTISAMPLE_DEPTH_OF_FIELD, m_multisample_depth_of_field);
        rt::check_menu(menu, IDM_MULTISAMPLE_MOTION_BLUR, m_multisample_motion_blur);
        break;

    case IDM_MULTISAMPLE_DEPTH_OF_FIELD:
        m_multisample_antialias = false;
        m_multisample_depth_of_field = !m_multisample_depth_of_field;
        m_multisample_motion_blur = false;
        rt::check_menu(menu, IDM_MULTISAMPLE_ANTIALIAS, m_multisample_antialias);
        rt::check_menu(menu, IDM_MULTISAMPLE_DEPTH_OF_FIELD, m_multisample_depth_of_field);
        rt::check_menu(menu, IDM_MULTISAMPLE_MOTION_BLUR, m_multisample_motion_blur);
        break;

    case IDM_MULTISAMPLE_MOTION_BLUR:
        m_multisample_antialias = false;
        m_multisample_depth_of_field = false;
        m_multisample_motion_blur = !m_multisample_motion_blur;
        rt::check_menu(menu, IDM_MULTISAMPLE_ANTIALIAS, m_multisample_antialias);
        rt::check_menu(menu, IDM_MULTISAMPLE_DEPTH_OF_FIELD, m_multisample_depth_of_field);
        rt::check_menu(menu, IDM_MULTISAMPLE_MOTION_BLUR, m_multisample_motion_blur);
        break;

    case IDM_BLEND_CLEAR:
    case IDM_BLEND_SRC:
    case IDM_BLEND_DEST:
    case IDM_BLEND_SRC_OVER_DEST:
    case IDM_BLEND_DEST_OVER_SRC:
    case IDM_BLEND_SRC_IN_DEST:
    case IDM_BLEND_DEST_IN_SRC:
    case IDM_BLEND_SRC_ATOP_DEST:
    case IDM_BLEND_DEST_ATOP_SRC:
    case IDM_BLEND_SRC_OUT_DEST:
    case IDM_BLEND_DEST_OUT_SRC:
    case IDM_BLEND_SRC_XOR_DEST:
    case IDM_BLEND_NONE:
        rt::check_menu(menu, IDM_BLEND_CLEAR + m_blending, false);
        m_blending = e_blending_operator(control - IDM_BLEND_CLEAR);
        rt::check_menu(menu, control, true);
        break;

    case IDM_RESET_VIEW:
        m_fWorldRotX = 0.0f;
        m_fWorldRotY = 0.0f;
        break;

    default:
        // all our control IDs are > 40006 and we should handle them all
        if (control > 40006)
        {
            ATLASSERT(false);
        }
    }
}

//-----------------------------------------------------------------------------
// CMyD3DApplication::on_texture_file
//
// Get a texture filename from the user and try to open it as the current
// texture.
//
void
CMyD3DApplication::on_texture_file(HWND window, WPARAM wp, LPARAM lp)
{
    TCHAR file[MAX_PATH] = { 0 };
    OPENFILENAME ofn =
    {
        sizeof(ofn), 0, 0,
            _T("All images\0")
                _T("*.bmp;*.jpg;*.png;*.pbm;*.pgm;*.ppm;*.pnm;*.tga;*.tif\0")
            _T("All files (*.*)\0*.*\0")
            _T("Bitmap images (*.bmp)\0*.bmp\0")
            _T("JPEG images (*.jpg)\0*.jpg\0")
            _T("PNG images (*.png)\0*.png\0")
            _T("Portable bitmap (*.pbm)\0*.pbm\0")
            _T("Portable graymap (*.pgm)\0*.pgm\0")
            _T("Portable pixmap (*.ppm)\0*.ppm\0")
            _T("Portable anymap (*.pnm)\0*.pnm\0")
            _T("Targa images(*.tga)\0*.tga\0")
            _T("TIFF images (*.tif)\0*.tif\0")
            _T("\0"),
        0,  0, 0,
        file, MAX_PATH, 0,  0, 0, 0, OFN_FILEMUSTEXIST
    };
    if (::GetOpenFileName(&ofn))
    {
        create_texture(file);
    }
}

//-----------------------------------------------------------------------------
// CMyD3DApplication::create_texture
//
// Creates the texture map from the given file.  If that was successful, it
// becomes the current texture used on the next frame draw.
//
void
CMyD3DApplication::create_texture(LPCTSTR file)
{
    CComPtr<IDirect3DTexture8> new_texture;

    const HRESULT hr = ::D3DUtil_CreateTexture(m_pd3dDevice,
            file ? file : m_texture_file.c_str(),
            &new_texture, D3DFMT_A8R8G8B8);
    if (SUCCEEDED(hr))
    {
        m_texture = new_texture;
        if (file)
        {
            m_texture_file = file;
        }
    }
    else
    {
        THR(hr);
    }
}

//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::tile_background
//
// Use CopyRects to tile the back buffer with an image containing alpha.
// The rectangle and point lists are rebuilt whenever RestoreDeviceObjects
// is called.  This keeps the lists updated whenever the device changes
// or the window is resized.
//
void
CMyD3DApplication::tile_background()
{
    // draw background tile, storing alpha straight into destination
    CComPtr<IDirect3DSurface8> back;
    THR(m_pd3dDevice->GetBackBuffer(0, D3DBACKBUFFER_TYPE_MONO, &back));
    THR(m_pd3dDevice->CopyRects(m_device_tile, &m_tile_rects[0],
        m_tile_rects.size(), back, &m_tile_offsets[0]));
}
