#if !defined(RT_TEXTURE_H)
#define RT_TEXTURE_H
//------------------------------------------------------------
// texture.h
//
// Helper functions for manipulating textures.
//
#include <atlbase.h>    // ATLASSERT(), CComPtr<>
#include <d3d8.h>       // IDirect3DTexture8

namespace rt {

//------------------------------------------------------------
// texture_lock
//
// Helper class that locks a texture in its c'tor and unlocks
// it in its d'tor.  Provides accessors to the locked region.
//
class texture_lock
{
private:
    // Pitch is specified in bytes, not pixels.
    const BYTE *scanline(UINT y) const
    {
        return static_cast<const BYTE *>(m_data.pBits)
            + m_data.Pitch*y;
    }
    BYTE *scanline(UINT y)
    {
        return static_cast<BYTE *>(m_data.pBits) +
            m_data.Pitch*y;
    }

    CComPtr<IDirect3DTexture8> m_texture;
    UINT m_level;
    D3DLOCKED_RECT m_data;

public:
    texture_lock(IDirect3DTexture8 *texture,
        UINT level = 0,
        DWORD flags = 0,
        const RECT *locked_region = NULL)
        : m_texture(texture),
        m_level(level)
    {
        THR(m_texture->LockRect(level, &m_data,
            locked_region, flags));
    }
    ~texture_lock()
    {
        // destructors should never throw exceptions, so
        // we don't use THR() here.  Also, we will never
        // be here unless the LockRect succeeded and
        // constructed a texture_lock, so UnlockRect
        // should always succeed.  We check anyway by
        // asserting success on the returned HRESULT.
        //
        // ATLASSERT is compiled out on optimized builds,
        // so use two statements because this:
        //      ATLASSERT(SUCCEEDED(m_texture->Unlock()))
        // would compile away the Unlock and introduce
        // a bug on a release build.
        const HRESULT hr = m_texture->UnlockRect(m_level);
        ATLASSERT(SUCCEEDED(hr));
    }

    // 8 bits per pixel: 1 pixel = 1 BYTE
    const BYTE *scanline8(UINT y) const
    {
        return scanline(y);
    }
    BYTE *scanline8(UINT y)
    {
        return scanline(y);
    }

    // 16 bits per pixel: 1 pixel = 1 WORD
    const WORD *scanline16(UINT y) const
    {
        return reinterpret_cast<const WORD *>(scanline(y));
    }
    WORD *scanline16(UINT y)
    {
        return reinterpret_cast<WORD *>(scanline(y));
    }

    // 24 bits per pixel: 1 pixel = 3 BYTEs
    const BYTE *scanline24(UINT y) const
    {
        return scanline(y);
    }
    BYTE *scanline24(UINT y)
    {
        return scanline(y);
    }

    // 32 bits per pixel: 1 pixel = 1 DWORD
    const DWORD *scanline32(UINT y) const
    {
        return reinterpret_cast<const DWORD *>(scanline(y));
    }
    DWORD *scanline32(UINT y)
    {
        return reinterpret_cast<DWORD *>(scanline(y));
    }
};  // texture_lock

};  // rt

#endif
