////////////////////////////////////////////////////////////
// rt_FrameBuffer.cpp
//
// Demonstrates frame buffer processing portions of the
// pipeline: stencil test, alpha blending, dithering,
// multisampling, and write masks.
//
#include <math.h>

#define STRICT
#include <windows.h>
#include <tchar.h>

#include <D3DX8.h>

#include "D3DUtil.h"
#include "resource.h"
#include "rt_FrameBuffer.h"

#include "rt/app.h"
#include "rt/hr.h"
#include "rt/mat.h"
#include "rt/misc.h"
#include "rt/states.h"

////////////////////////////////////////////////////////////
// determine blend factors based on blending operator
const struct s_blend_factor g_blend_factors[12] =
{
    // in order of e_blending_operator:
    // E_BLENDING_CLEAR
    D3DBLEND_ZERO,          D3DBLEND_ZERO,
    // E_BLENDING_SRC
    D3DBLEND_ONE,           D3DBLEND_ZERO,
    // E_BLENDING_DEST
    D3DBLEND_ZERO,          D3DBLEND_ONE,
    // E_BLENDING_SRC_OVER_DEST
    D3DBLEND_ONE,           D3DBLEND_INVSRCALPHA,
    // E_BLENDING_DEST_OVER_SRC
    D3DBLEND_INVDESTALPHA,  D3DBLEND_ONE,
    // E_BLENDING_SRC_IN_DEST
    D3DBLEND_DESTALPHA,     D3DBLEND_ZERO,
    // E_BLENDING_DEST_IN_SRC
    D3DBLEND_ZERO,          D3DBLEND_SRCALPHA,
    // E_BLENDING_SRC_ATOP_DEST
    D3DBLEND_DESTALPHA,     D3DBLEND_INVSRCALPHA,
    // E_BLENDING_DEST_ATOP_SRC
    D3DBLEND_INVDESTALPHA,  D3DBLEND_SRCALPHA,
    // E_BLENDING_SRC_OUT_DEST
    D3DBLEND_INVDESTALPHA,  D3DBLEND_ZERO,
    // E_BLENDING_DEST_OUT_SRC
    D3DBLEND_ZERO,          D3DBLEND_INVSRCALPHA,
    // E_BLENDING_SRC_XOR_DEST
    D3DBLEND_INVDESTALPHA,  D3DBLEND_INVSRCALPHA
};

////////////////////////////////////////////////////////////
// set_color_write_enable
//
// Set RS Color Write Enable based on UI state.
//
void
CMyD3DApplication::set_color_write_enable()
{
    DWORD mask = 0;
    if (m_color_write_enable_red)
    {
        mask |= D3DCOLORWRITEENABLE_RED;
    }
    if (m_color_write_enable_green)
    {
        mask |= D3DCOLORWRITEENABLE_GREEN;
    }
    if (m_color_write_enable_blue)
    {
        mask |= D3DCOLORWRITEENABLE_BLUE;
    }
    if (m_color_write_enable_alpha)
    {
        mask |= D3DCOLORWRITEENABLE_ALPHA;
    }
    THR(m_pd3dDevice->SetRenderState(D3DRS_COLORWRITEENABLE,
        mask));
}

////////////////////////////////////////////////////////////
// render_stencil_mask
//
// Use the torus mesh to render a hollow circle mask into
// the 0x1 stencil plane.
//
void
CMyD3DApplication::render_stencil_mask()
{
    // set the reference and mask values for plane 1
    rt::s_rs states[] =
    {
        D3DRS_STENCILREF, 0x1,
        D3DRS_STENCILMASK, 0x1,
        D3DRS_STENCILWRITEMASK, 0x1,
    };
    rt::set_states(m_pd3dDevice, states, NUM_OF(states));
    // disable texturing
    const rt::s_tss ts_states[] =
    {
        D3DTSS_COLOROP, D3DTOP_DISABLE,
        D3DTSS_ALPHAOP, D3DTOP_DISABLE
    };
    rt::set_states(m_pd3dDevice, 0, ts_states,
        NUM_OF(ts_states));
    THR(m_pd3dDevice->SetTexture(0, NULL));

    // draw torus
    D3DXMATRIX ident(1, 0, 0, 0,
                     0, 1, 0, 0,
                     0, 0, 1, 0,
                     0, 0, 0, 1);
    THR(m_pd3dDevice->SetTransform(D3DTS_WORLD, &ident));
    THR(m_torus->DrawSubset(0));
}

////////////////////////////////////////////////////////////
// render_stencil_stipple
//
// Render a stipple pattern into the 0x2 stencil plane.
//
void
CMyD3DApplication::render_stencil_stipple()
{
    // set the reference and mask values for plane 2; the
    // alpha test rejects stipple texels with zero alpha,
    // so that the stencil bits are not written there.
    rt::s_rs states[] =
    {
        D3DRS_ALPHATESTENABLE, true,
        D3DRS_ALPHAFUNC, D3DCMP_GREATEREQUAL,
        D3DRS_ALPHAREF, 0x1,
        D3DRS_STENCILREF, 0x2,
        D3DRS_STENCILMASK, 0x2,
        D3DRS_STENCILWRITEMASK, 0x2,
        D3DRS_COLORVERTEX, true,
        D3DRS_LIGHTING, false,
        D3DRS_SPECULARENABLE, false
    };
    rt::set_states(m_pd3dDevice, states, NUM_OF(states));
    rt::s_tss tex_states[] =
    {
        D3DTSS_COLOROP, D3DTOP_SELECTARG1,
        D3DTSS_COLORARG1, D3DTA_TEXTURE,
        D3DTSS_COLORARG2, D3DTA_DIFFUSE,
        D3DTSS_ALPHAOP, D3DTOP_SELECTARG1,
        D3DTSS_ALPHAARG1, D3DTA_TEXTURE,
        D3DTSS_ALPHAARG2, D3DTA_DIFFUSE
    };
    rt::set_states(m_pd3dDevice, 0,
        tex_states, NUM_OF(tex_states));
    THR(m_pd3dDevice->SetTexture(0, m_stipple));

    // draw stipple with screen-space triangles that cover
    // the whole screen
    THR(m_pd3dDevice->SetVertexShader(s_screen_vertex::FVF));
    THR(m_pd3dDevice->SetStreamSource(0, m_stipple_verts,
        sizeof(s_screen_vertex)));
    THR(m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0,
        m_num_stipple_quads*2));
}

////////////////////////////////////////////////////////////
// render_stencil
//
// Build the stencil plane(s) depending on UI state.
//
void
CMyD3DApplication::render_stencil()
{
    // clear the stencil planes first
    THR(m_pd3dDevice->Clear(0L, NULL, D3DCLEAR_STENCIL,
        0, 1.0f, 0L));

    // set states for stencil rendering
    rt::s_rs states[] =
    {
        D3DRS_STENCILENABLE, true,
        D3DRS_STENCILFUNC, D3DCMP_ALWAYS,
        D3DRS_STENCILPASS, D3DSTENCILOP_REPLACE,
        D3DRS_STENCILFAIL, D3DSTENCILOP_REPLACE,
        D3DRS_STENCILZFAIL, D3DSTENCILOP_REPLACE,
        D3DRS_COLORWRITEENABLE, 0L,
        D3DRS_ALPHABLENDENABLE, true,
        D3DRS_SRCBLEND, D3DBLEND_ZERO,
        D3DRS_DESTBLEND, D3DBLEND_ONE
    };
    rt::set_states(m_pd3dDevice, states, NUM_OF(states));

    // render geometry to set the planes
    if (m_stencil_mask)
    {
        render_stencil_mask();
    }
    if (m_stencil_stippling)
    {
        render_stencil_stipple();
    }

    // set the state for masking against stencil bits
    rt::s_rs use_stencil_states[] =
    {
        D3DRS_STENCILENABLE, true,
        D3DRS_STENCILFUNC, D3DCMP_EQUAL,
        D3DRS_STENCILREF, 0,
        D3DRS_STENCILMASK, ~0,
        D3DRS_STENCILWRITEMASK, 0,
        D3DRS_STENCILPASS, D3DSTENCILOP_KEEP,
        D3DRS_STENCILFAIL, D3DSTENCILOP_KEEP,
        D3DRS_STENCILZFAIL, D3DSTENCILOP_KEEP
    };
    rt::set_states(m_pd3dDevice, use_stencil_states,
        NUM_OF(use_stencil_states));
    set_color_write_enable();
}

////////////////////////////////////////////////////////////
// render_teapot
//
// Render the teapot with a given world matrix.
//
void
CMyD3DApplication::render_teapot(const D3DXMATRIX &matrix)
{
    THR(m_pd3dDevice->SetTransform(D3DTS_WORLD, &matrix));
    THR(m_teapot->DrawSubset(0));
}

////////////////////////////////////////////////////////////
// render_motion_blur
//
// Render multiple copies of the teapot, changing the
// multisample mask to create a motion blur effect.
//
void
CMyD3DApplication::render_motion_blur()
{
    UINT passes = max(UINT(m_d3dpp.MultiSampleType), 1);
    float rot_x = m_last_rot_x;
    float delta_x = (m_fWorldRotX - m_last_rot_x)/passes;
    float rot_y = m_last_rot_y;
    float delta_y = (m_fWorldRotY - m_last_rot_y)/passes;

    for (UINT i = 0; i < passes; i++)
    {
        THR(m_pd3dDevice->SetRenderState(
            D3DRS_MULTISAMPLEMASK, 1L << i));
        render_teapot(rt::mat_rot_x(rot_x)*
                      rt::mat_rot_y(rot_y));
        rot_x += delta_x;
        rot_y += delta_y;
    }
    THR(m_pd3dDevice->SetRenderState(D3DRS_MULTISAMPLEMASK,
        ~0));
}

////////////////////////////////////////////////////////////
// render_depth_of_field
//
// Render 9 teapots with a jittered view frustum to
// create a depth of field effect.
//
void
CMyD3DApplication::render_depth_of_field()
{
    const D3DXMATRIX base = rt::mat_rot_x(m_fWorldRotX)*
        rt::mat_rot_y(m_fWorldRotY)*rt::mat_scale(0.5f);

    const UINT passes =
        max(UINT(m_d3dpp.MultiSampleType), 1);
    const float *jitter = g_jitter[passes-1];
    for (UINT i = 0; i < passes; i++)
    {
        THR(m_pd3dDevice->SetRenderState(
            D3DRS_MULTISAMPLEMASK, 1L << i));

        rt::mat_look_at view(D3DXVECTOR3(
            jitter[i*2+0]*0.25f, jitter[i*2+1]*0.25f, -5));
        THR(m_pd3dDevice->SetTransform(D3DTS_VIEW, &view));

        render_teapot(base*rt::mat_trans(-1, -1, 1));
        render_teapot(base*rt::mat_trans(1, -1, 1));
        render_teapot(base*rt::mat_trans(-1, 1, 1));
        render_teapot(base*rt::mat_trans(1, 1, 1));
        render_teapot(base);
        render_teapot(base*rt::mat_trans(-1, -1, -1));
        render_teapot(base*rt::mat_trans(1, -1, -1));
        render_teapot(base*rt::mat_trans(-1, 1, -1));
        render_teapot(base*rt::mat_trans(1, 1, -1));
    }

    rt::mat_look_at view(D3DXVECTOR3(0, 0, -5));
    THR(m_pd3dDevice->SetTransform(D3DTS_VIEW, &view));
    THR(m_pd3dDevice->SetRenderState(D3DRS_MULTISAMPLEMASK,
        ~0));
}

////////////////////////////////////////////////////////////
// Render()
//
// Render a frame consisting of these elements:
// - Background color
// - Tiled background image via CopyRects
// - Stencil mask or stipple, if needed
// - Texturing states
// - Blending states
// - Blended, textured teapot
// - Text overlay, if needed
//
HRESULT CMyD3DApplication::Render()
{
    // Clear the viewport
    THR(m_pd3dDevice->Clear(0L, NULL, D3DCLEAR_TARGET |
        D3DCLEAR_ZBUFFER, m_bg, 1.0f, 0L));

    if (m_tile_background)
    {
        tile_background();
    }

    // enable or disable full-scene antialiasing
    THR(m_pd3dDevice->SetRenderState(
        D3DRS_MULTISAMPLEANTIALIAS,
        m_multisample_antialias));

    THR(m_pd3dDevice->BeginScene());

    // set up state for rendering: materials & lighting
    D3DMATERIAL8 mtrl;
    ::D3DUtil_InitMaterial(mtrl, 1.0f, 0.0f, 0.0f);
    mtrl.Diffuse = D3DXCOLOR(m_fg);
    mtrl.Specular = D3DXCOLOR(m_specular);
    THR(m_pd3dDevice->SetMaterial(&mtrl));

    // build stencil planes, if necessary
    if (m_build_stencil)
    {
        render_stencil();
        m_build_stencil = false;
    }
        
    // texturing states
    if (m_can_texture && m_textured)
    {
        const rt::s_tss ts_states[] =
        {
            D3DTSS_TEXCOORDINDEX,
                D3DTSS_TCI_CAMERASPACEPOSITION,
            D3DTSS_COLORARG1, D3DTA_TEXTURE,
            D3DTSS_COLOROP, D3DTOP_MODULATE,
            D3DTSS_COLORARG2, D3DTA_DIFFUSE,
            D3DTSS_ALPHAARG1, D3DTA_TEXTURE,
            D3DTSS_ALPHAOP, D3DTOP_SELECTARG2,
            D3DTSS_ALPHAARG2, D3DTA_DIFFUSE,
            D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2,
            D3DTSS_ADDRESSU, D3DTADDRESS_WRAP,
            D3DTSS_ADDRESSV, D3DTADDRESS_WRAP,
            D3DTSS_MINFILTER, D3DTEXF_LINEAR,
            D3DTSS_MAGFILTER, D3DTEXF_LINEAR,
            D3DTSS_MIPFILTER, D3DTEXF_LINEAR
        };
        rt::set_states(m_pd3dDevice, 0, ts_states,
            NUM_OF(ts_states));
        ATLASSERT(m_texture);
        THR(m_pd3dDevice->SetTexture(0, m_texture));
        // set a (u,v) transform that makes the flag
        // rectangular and positioned nicely on the
        // teapot.  Invert the y-axis to get it in
        // the proper vertical orientation.  Rotate
        // it around the Z axis to make it a little
        // more interesting.
        D3DXMATRIX texform =
            rt::mat_rot_z(D3DX_PI/4.0f)*
            rt::mat_scale(0.5f, -1, 1)*
            rt::mat_trans(-0.6f, -0.5f, 0);
        THR(m_pd3dDevice->SetTransform(D3DTS_TEXTURE0,
            &texform));
    }
    else
    {
        // disable texturing
        const rt::s_tss ts_states[] =
        {
            D3DTSS_TEXCOORDINDEX, 0,
            D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE,
            D3DTSS_COLOROP, D3DTOP_DISABLE,
            D3DTSS_ALPHAOP, D3DTOP_DISABLE
        };
        rt::set_states(m_pd3dDevice, 0, ts_states,
            NUM_OF(ts_states));
        THR(m_pd3dDevice->SetTexture(0, NULL));
    }

    // stencil test state
    if (m_stencil_mask || m_stencil_stippling)
    {
        // draw teapot only where no stencil bits set,
        // and don't modify the stencil bits
        rt::s_rs states[] =
        {
            D3DRS_STENCILENABLE, true,
            D3DRS_STENCILFUNC, D3DCMP_EQUAL,
            D3DRS_STENCILREF, 0,
            D3DRS_STENCILMASK, ~0,
            D3DRS_STENCILWRITEMASK, 0,
            D3DRS_STENCILFAIL, D3DSTENCILOP_KEEP,
            D3DRS_STENCILPASS, D3DSTENCILOP_KEEP,
            D3DRS_STENCILZFAIL, D3DSTENCILOP_KEEP
        };
        rt::set_states(m_pd3dDevice, states,
            NUM_OF(states));
    }
    else
    {
        THR(m_pd3dDevice->SetRenderState(
            D3DRS_STENCILENABLE, false));
    }

    // setup blending state
    if (E_BLENDING_NONE == m_blending)
    {
        THR(m_pd3dDevice->SetRenderState(
            D3DRS_ALPHABLENDENABLE, false));
    }
    else
    {
        rt::s_rs states[] =
        {
            D3DRS_ALPHABLENDENABLE, true,
            D3DRS_SRCBLEND,
                g_blend_factors[m_blending].m_src,
            D3DRS_DESTBLEND,
                g_blend_factors[m_blending].m_dest,
            D3DRS_BLENDOP, D3DBLENDOP_ADD
        };
        rt::set_states(m_pd3dDevice, states,
            NUM_OF(states));
    }

    // frame buffer render states
    const rt::s_rs states[] =
    {
        D3DRS_NORMALIZENORMALS, true,
        D3DRS_LIGHTING, true,
        D3DRS_SPECULARENABLE, true,
        D3DRS_ZENABLE, true,
        D3DRS_ZFUNC, D3DCMP_LESS,
        D3DRS_ZBIAS, 0,
        D3DRS_DITHERENABLE, m_dithered,
        D3DRS_ZWRITEENABLE, m_z_write_enable
    };
    rt::set_states(m_pd3dDevice, states, NUM_OF(states));

    // Render the teapot mesh
    if (m_multisample_depth_of_field)
    {
        render_depth_of_field();
    }
    else if (m_multisample_motion_blur)
    {
        render_motion_blur();
    }
    else
    {
        render_teapot(rt::mat_rot_x(m_fWorldRotX)*
                      rt::mat_rot_y(m_fWorldRotY));
    }

    // Render stats and help text
    if (m_show_stats)
    {
        RenderText();
    }

    THR(m_pd3dDevice->EndScene());

    return S_OK;
}
